{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# User Guide\n",
    "This user guide walks through the basics of reading and writing data with the Ray Snowflake connector.\n",
    "\n",
    "## Before you begin\n",
    "- [Install](install.ipynb) the connector.\n",
    "- [Define](connection.ipynb) connection properties.\n",
    "- [Setup](setup.ipynb) database objects. \n",
    "\n",
    "## How it works\n",
    "The Ray connector for Snowflake holds the connection properties required to establish a connection with Snowflake. The connector implements the same interface as a DBAPi2Connector, with the normal `query` methods as well as `read_table` and `write_table` methods optimized for large scale parallel data exchange with Snowflake. \n",
    "\n",
    "### Connector factory\n",
    "The `SnowflakeConnector` class has a factory `create` method that is used to create the connector. The `create` method can be called with any connection properties, as well as an optional connection property file path. Read more [here](connection.ipynb) about how connection properties are read.\n",
    "\n",
    "### Connections and transactions\n",
    "The connector is a Python context manage, and can be used the `with` semantics to define when a connection should be established, db operations commited to the database, and the connection closed. \n",
    "\n",
    "![Transaction context](../../images/snowflake_transactions.png)\n",
    "\n",
    "The code below will read from sample table using the connector to manage the connection."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "150000\n"
     ]
    }
   ],
   "source": [
    "\n",
    "from ray_db.provider.snowflake import SnowflakeConnector\n",
    "import ray\n",
    "\n",
    "if not ray.is_initialized():\n",
    "    ray.init(logging_level='ERROR', runtime_env={'env_vars':{'RAY_DISABLE_PYARROW_VERSION_CHECK': '1'}})\n",
    "\n",
    "SOURCE_TABLE = 'CUSTOMER'\n",
    "\n",
    "connector = SnowflakeConnector.create()\n",
    "with connector:\n",
    "    count = connector.query_value(f'SELECT COUNT(*) from {SOURCE_TABLE}')\n",
    "    \n",
    "print(count)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Alternatively, you can use `try` blocks with the connector's `open`, `commit` and `close` methods. "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "150000\n"
     ]
    }
   ],
   "source": [
    "try:\n",
    "    connector.open()\n",
    "    count = connector.query_value(f'SELECT COUNT(*) from {SOURCE_TABLE}')\n",
    "    \n",
    "    # if this was a write operation:\n",
    "    # connector.commit()\n",
    "finally:\n",
    "    connector.close()\n",
    "    \n",
    "print(count)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Read from tables\n",
    "In order to read an entire table into a a Ray cluster, the connector provides a `read_table` method. Under the hood, Ray will use Snowflake optimizations that allow query results to be read in parallel. This enable the connector to read in gigbytes of data per second into a Ray cluster. The Snowflake API has a `read_batch` method that enables this optimization. Ray datasets are composed of PyArrow tables that are spread across the entrire Ray cluster to allow for the distributed operations required in machine learning.\n",
    "\n",
    "![Snowflake read table](../../images/snowflake_read_table.png)\n",
    "\n",
    "The code below will read in a sample table from the Snowflake sample database. The `fully_executed` method triggers the entire dataset to load. If this is not called, the data is only loaded when a dataset operation requires it."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "ename": "RayTaskError(ImportError)",
     "evalue": "\u001b[36mray::_get_read_tasks()\u001b[39m (pid=95812, ip=10.0.21.126)\n  File \"/home/ray/anaconda3/lib/python3.10/site-packages/ray/data/read_api.py\", line 1380, in _get_read_tasks\n    reader = ds.create_reader(**kwargs)\n  File \"/tmp/ray/session_2022-11-15_12-34-06_679234_139/runtime_resources/working_dir_files/_ray_pkg_bfb55b36b2eeb9b7cb1dd2cfbf4abe5c/ray_db/__init__.py\", line 207, in create_reader\n    self.reader.prepare(**read_args)\n  File \"/tmp/ray/session_2022-11-15_12-34-06_679234_139/runtime_resources/working_dir_files/_ray_pkg_bfb55b36b2eeb9b7cb1dd2cfbf4abe5c/ray_db/provider/dbapi2/__init__.py\", line 326, in prepare\n    self.sample = BlockAccessor.for_block(sample).to_arrow()\n  File \"/home/ray/anaconda3/lib/python3.10/site-packages/ray/data/block.py\", line 378, in for_block\n    _check_pyarrow_version()\n  File \"/home/ray/anaconda3/lib/python3.10/site-packages/ray/data/_internal/util.py\", line 78, in _check_pyarrow_version\n    raise ImportError(\nImportError: Datasets requires pyarrow >= 6.0.1, < 7.0.0, but 8.0.0 is installed. Reinstall with `pip install -U \"pyarrow<7.0.0\"`. If you want to disable this pyarrow version check, set the environment variable RAY_DISABLE_PYARROW_VERSION_CHECK=1.",
     "output_type": "error",
     "traceback": [
      "\u001b[0;31m---------------------------------------------------------------------------\u001b[0m",
      "\u001b[0;31mRayTaskError(ImportError)\u001b[0m                 Traceback (most recent call last)",
      "Cell \u001b[0;32mIn [9], line 2\u001b[0m\n\u001b[1;32m      1\u001b[0m \u001b[39mwith\u001b[39;00m connector:\n\u001b[0;32m----> 2\u001b[0m     ds \u001b[39m=\u001b[39m connector\u001b[39m.\u001b[39;49mread_table(SOURCE_TABLE)\u001b[39m.\u001b[39mfully_executed()\n\u001b[1;32m      4\u001b[0m ds\u001b[39m.\u001b[39mlimit(\u001b[39m3\u001b[39m)\u001b[39m.\u001b[39mto_pandas()\n",
      "File \u001b[0;32m~/egr-workspace/ray_db/provider/snowflake/__init__.py:124\u001b[0m, in \u001b[0;36mSnowflakeConnector.read_table\u001b[0;34m(self, table, columns, **kwargs)\u001b[0m\n\u001b[1;32m    122\u001b[0m writer \u001b[39m=\u001b[39m WriteSnowflakeStages(\u001b[39mself\u001b[39m, table, columns)  \n\u001b[1;32m    123\u001b[0m datasource \u001b[39m=\u001b[39m DatabaseTableDatasource(reader, writer)\n\u001b[0;32m--> 124\u001b[0m \u001b[39mreturn\u001b[39;00m ray\u001b[39m.\u001b[39;49mdata\u001b[39m.\u001b[39;49mread_datasource(datasource, \u001b[39m*\u001b[39;49m\u001b[39m*\u001b[39;49mkwargs)\n",
      "File \u001b[0;32m~/anaconda3/lib/python3.10/site-packages/ray/data/read_api.py:273\u001b[0m, in \u001b[0;36mread_datasource\u001b[0;34m(datasource, parallelism, ray_remote_args, **read_args)\u001b[0m\n\u001b[1;32m    266\u001b[0m \u001b[39melse\u001b[39;00m:\n\u001b[1;32m    267\u001b[0m     \u001b[39m# Prepare read in a remote task so that in Ray client mode, we aren't\u001b[39;00m\n\u001b[1;32m    268\u001b[0m     \u001b[39m# attempting metadata resolution from the client machine.\u001b[39;00m\n\u001b[1;32m    269\u001b[0m     get_read_tasks \u001b[39m=\u001b[39m cached_remote_fn(\n\u001b[1;32m    270\u001b[0m         _get_read_tasks, retry_exceptions\u001b[39m=\u001b[39m\u001b[39mFalse\u001b[39;00m, num_cpus\u001b[39m=\u001b[39m\u001b[39m0\u001b[39m\n\u001b[1;32m    271\u001b[0m     )\n\u001b[0;32m--> 273\u001b[0m     requested_parallelism, min_safe_parallelism, read_tasks \u001b[39m=\u001b[39m ray\u001b[39m.\u001b[39;49mget(\n\u001b[1;32m    274\u001b[0m         get_read_tasks\u001b[39m.\u001b[39;49mremote(\n\u001b[1;32m    275\u001b[0m             datasource,\n\u001b[1;32m    276\u001b[0m             ctx,\n\u001b[1;32m    277\u001b[0m             cur_pg,\n\u001b[1;32m    278\u001b[0m             parallelism,\n\u001b[1;32m    279\u001b[0m             _wrap_and_register_arrow_serialization_workaround(read_args),\n\u001b[1;32m    280\u001b[0m         )\n\u001b[1;32m    281\u001b[0m     )\n\u001b[1;32m    283\u001b[0m \u001b[39mif\u001b[39;00m read_tasks \u001b[39mand\u001b[39;00m \u001b[39mlen\u001b[39m(read_tasks) \u001b[39m<\u001b[39m min_safe_parallelism \u001b[39m*\u001b[39m \u001b[39m0.7\u001b[39m:\n\u001b[1;32m    284\u001b[0m     perc \u001b[39m=\u001b[39m \u001b[39m1\u001b[39m \u001b[39m+\u001b[39m \u001b[39mround\u001b[39m((min_safe_parallelism \u001b[39m-\u001b[39m \u001b[39mlen\u001b[39m(read_tasks)) \u001b[39m/\u001b[39m \u001b[39mlen\u001b[39m(read_tasks), \u001b[39m1\u001b[39m)\n",
      "File \u001b[0;32m~/anaconda3/lib/python3.10/site-packages/ray/_private/client_mode_hook.py:105\u001b[0m, in \u001b[0;36mclient_mode_hook.<locals>.wrapper\u001b[0;34m(*args, **kwargs)\u001b[0m\n\u001b[1;32m    103\u001b[0m     \u001b[39mif\u001b[39;00m func\u001b[39m.\u001b[39m\u001b[39m__name__\u001b[39m \u001b[39m!=\u001b[39m \u001b[39m\"\u001b[39m\u001b[39minit\u001b[39m\u001b[39m\"\u001b[39m \u001b[39mor\u001b[39;00m is_client_mode_enabled_by_default:\n\u001b[1;32m    104\u001b[0m         \u001b[39mreturn\u001b[39;00m \u001b[39mgetattr\u001b[39m(ray, func\u001b[39m.\u001b[39m\u001b[39m__name__\u001b[39m)(\u001b[39m*\u001b[39margs, \u001b[39m*\u001b[39m\u001b[39m*\u001b[39mkwargs)\n\u001b[0;32m--> 105\u001b[0m \u001b[39mreturn\u001b[39;00m func(\u001b[39m*\u001b[39;49margs, \u001b[39m*\u001b[39;49m\u001b[39m*\u001b[39;49mkwargs)\n",
      "File \u001b[0;32m~/anaconda3/lib/python3.10/site-packages/ray/_private/worker.py:2289\u001b[0m, in \u001b[0;36mget\u001b[0;34m(object_refs, timeout)\u001b[0m\n\u001b[1;32m   2287\u001b[0m     worker\u001b[39m.\u001b[39mcore_worker\u001b[39m.\u001b[39mdump_object_store_memory_usage()\n\u001b[1;32m   2288\u001b[0m \u001b[39mif\u001b[39;00m \u001b[39misinstance\u001b[39m(value, RayTaskError):\n\u001b[0;32m-> 2289\u001b[0m     \u001b[39mraise\u001b[39;00m value\u001b[39m.\u001b[39mas_instanceof_cause()\n\u001b[1;32m   2290\u001b[0m \u001b[39melse\u001b[39;00m:\n\u001b[1;32m   2291\u001b[0m     \u001b[39mraise\u001b[39;00m value\n",
      "\u001b[0;31mRayTaskError(ImportError)\u001b[0m: \u001b[36mray::_get_read_tasks()\u001b[39m (pid=95812, ip=10.0.21.126)\n  File \"/home/ray/anaconda3/lib/python3.10/site-packages/ray/data/read_api.py\", line 1380, in _get_read_tasks\n    reader = ds.create_reader(**kwargs)\n  File \"/tmp/ray/session_2022-11-15_12-34-06_679234_139/runtime_resources/working_dir_files/_ray_pkg_bfb55b36b2eeb9b7cb1dd2cfbf4abe5c/ray_db/__init__.py\", line 207, in create_reader\n    self.reader.prepare(**read_args)\n  File \"/tmp/ray/session_2022-11-15_12-34-06_679234_139/runtime_resources/working_dir_files/_ray_pkg_bfb55b36b2eeb9b7cb1dd2cfbf4abe5c/ray_db/provider/dbapi2/__init__.py\", line 326, in prepare\n    self.sample = BlockAccessor.for_block(sample).to_arrow()\n  File \"/home/ray/anaconda3/lib/python3.10/site-packages/ray/data/block.py\", line 378, in for_block\n    _check_pyarrow_version()\n  File \"/home/ray/anaconda3/lib/python3.10/site-packages/ray/data/_internal/util.py\", line 78, in _check_pyarrow_version\n    raise ImportError(\nImportError: Datasets requires pyarrow >= 6.0.1, < 7.0.0, but 8.0.0 is installed. Reinstall with `pip install -U \"pyarrow<7.0.0\"`. If you want to disable this pyarrow version check, set the environment variable RAY_DISABLE_PYARROW_VERSION_CHECK=1."
     ]
    }
   ],
   "source": [
    "with connector:\n",
    "    ds = connector.read_table(SOURCE_TABLE).fully_executed()\n",
    "        \n",
    "ds.limit(3).to_pandas()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## DML and DDL\n",
    "The connector can also be used for any DDL or DML operations you would normally execute through the Snowflake Python API. These operations just pass through to the underlying Snowflake API. \n",
    "\n",
    "The code below will create the objects needed for writing to tables. Note that a commit is issued between our queries so the DDL operation executes prior to the next one that is dependent. An alternative is to use two `with` blocks to define transaction boundaries."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "DEST_DATABASE = 'ANYSCALE_SAMPLE'\n",
    "DEST_TABLE = f'CUSTOMER_COPY'\n",
    "\n",
    "with connector:\n",
    "    connector.query(f'CREATE DATABASE IF NOT EXISTS {DEST_DATABASE}')\n",
    "    connector.commit()\n",
    "    connector.query(f'CREATE OR REPLACE TABLE {DEST_TABLE} LIKE {SOURCE_TABLE}')"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Write to tables\n",
    "In order to write a dataset into database table, the connector provides a `write_table` method. Under the hood, the connector will use `PUT` operations to copy the Ray datasets PyArrow tables to a Snowflake tables stage. This happens in parallel, so every worker within Ray that holds a PyArrow table for the dataset will write to Snowflake at the same time. After the parallel put operation, the connector will issue a single `COPY` command to Snowflake which will trigger the Snowflake warehouse to beign moving the raw data into the production table. Since this is divided into two operations, one in parallel, and the second as a single operation, this is a two phased commmit. If any of the intial put opertations fail, the final copy operation will fail. \n",
    "\n",
    "![Snowflake write table](../../images/snowflake_write_table.png)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The code below writes the previously read dataset into the newly created table."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "with connector:\n",
    "    connector.write_table(ds, DEST_TABLE)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "The table can now be read."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "with connector:\n",
    "    ds2 = connector.read_table(DEST_TABLE).fully_executed()\n",
    "\n",
    "ds2.limit(3).to_pandas()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3.10.4 ('base')",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.10.4"
  },
  "orig_nbformat": 4,
  "vscode": {
   "interpreter": {
    "hash": "5abf9a257024fa0ae177d32ddc0977bda32aa95f4f2d5d07f829679a9e9e7642"
   }
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
